import 'package:flutter/material.dart';
import 'package:flutter_hooks/flutter_hooks.dart';
import 'package:flutter_query/src/query.dart';
import 'package:provider/provider.dart';

AsyncSnapshot<T> useQuery<T>(String key, QueryFunction<T> fn) {
  return use(_QueryHook(key, fn));
}

class _QueryHook<T> extends Hook<AsyncSnapshot<T>> {
  const _QueryHook(this.key, this.fn);

  final String key;
  final QueryFunction<T> fn;

  @override
  _QueryStateHook<T> createState() => _QueryStateHook();
}

class _QueryStateHook<T> extends HookState<AsyncSnapshot<T>, _QueryHook<T>> {
  AsyncSnapshot<T> _snapshot;
  Function _watcher;
  QueryCache _cache;

  @override
  void initHook() {
    _snapshot = AsyncSnapshot<T>.withData(ConnectionState.none, null);

    try {
      _cache = Provider.of<QueryCache>(context, listen: false);
      final key = hook.key;
      final query = _cache.registerQuery(key, hook.fn);
      if (query.lastValue != null) {
        _snapshot =
            AsyncSnapshot<T>.withData(ConnectionState.done, query.lastValue);
      }
      _watcher = (data) {
        if (data != null) {
          setState(() {
            _snapshot = AsyncSnapshot<T>.withData(ConnectionState.done, data);
          });
        }
      };
      _cache.watchQuery(key, _watcher);
      _cache.invalidateQuery(key);
    } catch (e) {
      throw Exception(
        'Cache not found! You either need to use the QueryCacheProvider or provide a query cache yourself.',
      );
    }
    super.initHook();
  }

  @override
  void dispose() {
    _cache.unwatchQuery(hook.key, _watcher);
    super.dispose();
  }

  @override
  AsyncSnapshot<T> build(BuildContext context) {
    return _snapshot;
  }
}
