import 'dart:async';

import 'package:flutter/material.dart';
import 'package:provider/provider.dart';

typedef Future<T> QueryFunction<T>();

class Query<T> {
  final String key;
  QueryFunction<T> fn;
  List<Function> observers = [];
  T lastValue;

  Query(this.key, QueryFunction<T> fn) {
    this.fn = () {
      return fn().then((data) {
        lastValue = data;
        return data;
      });
    };
  }
}

class QueryCache {
  Map<String, Query> _queryMap = {};

  Query getQueryByKey(String key) {
    return _queryMap[key];
  }

  Query registerQuery(String key, QueryFunction fn) {
    var query = getQueryByKey(key);
    if (query == null) {
      query = Query(key, fn);
      _queryMap[key] = query;
    }
    return query;
  }

  void watchQuery(String key, Function observer) {
    var query = getQueryByKey(key);
    if (query != null) {
      query.observers.add(observer);
    }
  }

  void unwatchQuery(String key, Function observer) {
    var query = getQueryByKey(key);
    if (query != null) {
      query.observers.remove(observer);
    }
  }

  void invalidateQuery(String key) {
    final query = getQueryByKey(key);
    if (query != null) {
      query.fn().then((data) {
        query.observers.forEach((observer) => observer(data));
      });
    }
  }
}

class QueryCacheProvider extends StatefulWidget {
  final Widget child;
  QueryCacheProvider({this.child});

  @override
  State<StatefulWidget> createState() => QueryCacheProviderState();
}

class QueryCacheProviderState extends State<QueryCacheProvider> {
  QueryCache _queryCache;

  @override
  void initState() {
    _queryCache = QueryCache();
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Provider.value(
      value: _queryCache,
      child: widget.child,
    );
  }
}
